import os

from typing import List, Dict, Callable, Any, Optional
from email.message import EmailMessage
from email.headerregistry import Address

from ...intel.vulnerabilities import ScannableSoftware, Vulnerability, \
        Software, ProductionVulnerability
from ...api.intelligence import VulnerabilityFeedVariant
from ...util.terminal import Color, escape, RESET
from ...util.html import Tag
from ...util.versioning import version_to_str
from ..reporting import Report, ReportColumnEnum, ReportFormatEnum, \
        ReportRecord, ReportManager, ReportFormat, ReportColumn, \
        BaseHumanReadableWriter, ReportEmail, get_config_options, \
        generate_report_email_html, generate_html_table, \
        REPORT_FORMAT_CSV, REPORT_FORMAT_TSV, REPORT_FORMAT_NULL_DELIMITED, \
        REPORT_FORMAT_LINE_DELIMITED
from ..context import CliContext
from ..email import Mailer


class VulnScanReportColumn(ReportColumnEnum):
    SOFTWARE_TYPE = 'software_type', lambda record: record.software.type.value
    SLUG = 'slug', lambda record: record.software.slug
    VERSION = 'version', lambda record: version_to_str(record.software.version)
    ID = 'id', \
        lambda record: record.vulnerability.identifier
    TITLE = 'title', lambda record: record.vulnerability.title
    LINK = 'link', lambda record: record.vulnerability.get_wordfence_link()
    DESCRIPTION = 'description', \
        lambda record: record.vulnerability.description, \
        VulnerabilityFeedVariant.PRODUCTION
    CVE = 'cve', lambda record: record.vulnerability.cve, \
        VulnerabilityFeedVariant.PRODUCTION
    CVSS_VECTOR = 'cvss_vector', \
        lambda record: record.vulnerability.cvss.vector, \
        VulnerabilityFeedVariant.PRODUCTION
    CVSS_SCORE = 'cvss_score', \
        lambda record: record.vulnerability.cvss.score, \
        VulnerabilityFeedVariant.PRODUCTION
    CVSS_RATING = 'cvss_rating', \
        lambda record: record.vulnerability.cvss.rating, \
        VulnerabilityFeedVariant.PRODUCTION
    CWE_ID = 'cwe_id', \
        lambda record: record.vulnerability.cwe.identifier, \
        VulnerabilityFeedVariant.PRODUCTION
    CWE_NAME = 'cwe_name', \
        lambda record: record.vulnerability.cwe.name, \
        VulnerabilityFeedVariant.PRODUCTION
    CWE_DESCRIPTION = 'cwe_description', \
        lambda record: record.vulnerability.cwe.description, \
        VulnerabilityFeedVariant.PRODUCTION
    PATCHED = 'patched', \
        lambda record: record.get_matched_software().patched
    REMEDIATION = 'remediation', \
        lambda record: record.get_matched_software().remediation, \
        VulnerabilityFeedVariant.PRODUCTION,
    PUBLISHED = 'published', lambda record: record.vulnerability.published
    UPDATED = 'updated', \
        lambda record: record.vulnerability.updated, \
        VulnerabilityFeedVariant.PRODUCTION
    SCANNED_PATH = 'scanned_path', lambda record: os.fsdecode(
            record.software.scan_path
        )

    def __init__(
                self,
                header: str,
                extractor: Callable[[Any], str],
                feed_variant: Optional[VulnerabilityFeedVariant] = None
            ):
        super().__init__(header, extractor)
        self.feed_variant = feed_variant

    def is_compatible(
                self,
                variant: VulnerabilityFeedVariant
            ) -> bool:
        return self.feed_variant is None or \
                variant == self.feed_variant


class HumanReadableWriter(BaseHumanReadableWriter):

    def get_severity_color(self, severity: str) -> str:
        if severity == 'none' or severity == 'low':
            return escape(color=Color.WHITE, bold=True)
        if severity == 'high' or severity == 'critical':
            return escape(color=Color.RED, bold=True)
        return escape(color=Color.YELLOW, bold=True)

    def format_record(self, record) -> str:
        vuln = record.vulnerability
        sw = record.software
        sw_version = version_to_str(sw.version)
        yellow = escape(color=Color.YELLOW)
        link = vuln.get_wordfence_link()
        blue = escape(color=Color.BLUE)
        white = escape(color=Color.WHITE)
        severity = None
        if isinstance(record.vulnerability, ProductionVulnerability):
            if record.vulnerability.cvss is not None:
                severity = record.vulnerability.cvss.rating
        if severity is None:
            severity_message = ''
        else:
            severity = severity.lower()
            severity_color = self.get_severity_color(severity)
            severity_message = f'{severity_color}{severity}{yellow} severity '
        if vuln.informational:
            bold_white = escape(color=Color.WHITE, bold=True)
            info_message = f'{bold_white}informational{yellow} '
            if len(severity_message) == 0:
                info_message = ' ' + info_message
        else:
            info_message = ''
        return (
            f'{yellow}Found {severity_message}{info_message}vulnerability '
            f'{vuln.title} in {sw.slug}({sw_version})\n'
            f'{white}Details: {blue}{link}{RESET}'
            )


REPORT_FORMAT_HUMAN = ReportFormat(
        'human',
        lambda stream, columns: HumanReadableWriter(stream),
        allows_headers=False,
        allows_column_customization=False
    )


class VulnScanReportFormat(ReportFormatEnum):
    CSV = REPORT_FORMAT_CSV
    TSV = REPORT_FORMAT_TSV
    NULL_DELIMITED = REPORT_FORMAT_NULL_DELIMITED
    LINE_DELIMITED = REPORT_FORMAT_LINE_DELIMITED
    HUMAN = REPORT_FORMAT_HUMAN


class VulnScanReportRecord(ReportRecord):

    def __init__(
                self,
                software: ScannableSoftware,
                vulnerability: Vulnerability
            ):
        self.software = software
        self.vulnerability = vulnerability
        self.matched_software = None

    def get_matched_software(self) -> Software:
        if self.matched_software is None:
            self.matched_software = \
                    self.vulnerability.get_matched_software(self.software)
        return self.matched_software


class VulnScanReport(Report):

    def __init__(
                self,
                format: VulnScanReportFormat,
                columns: List[VulnScanReportColumn],
                email_addresses: List[str],
                mailer: Optional[Mailer],
                write_headers: bool = False
            ):
        super().__init__(
                format=format,
                columns=columns,
                email_addresses=email_addresses,
                mailer=mailer,
                write_headers=write_headers
            )
        self.scanner = None

    def add_result(
                self,
                software: ScannableSoftware,
                vulnerabilities: Dict[str, Vulnerability]
            ) -> None:
        records = []
        for vulnerability in vulnerabilities.values():
            record = VulnScanReportRecord(
                    software,
                    vulnerability
                )
            records.append(record)
        self.write_records(records)

    def generate_email(
                self,
                recipient: Address,
                attachments: Dict[str, str],
                hostname: str
            ) -> EmailMessage:

        unique_count = self.scanner.get_vulnerability_count()
        total_count = self.scanner.get_total_count()

        base_message = 'Vulnerabilities were found by Wordfence CLI during ' \
                       'a scan.'

        plain = f'{base_message}\n\n' \
                f'Unique Vulnerabilities: {unique_count}\n' \
                f'Total Vulnerabilities: {total_count}\n'

        content = Tag('div')

        content.append(Tag('p').append(base_message))

        results = {
                'Unique Vulnerabilities': unique_count,
                'Total Vulnerabilities': total_count
            }
        table = generate_html_table(results)
        content.append(table)

        document = generate_report_email_html(
                content,
                'Vulnerability Scan Results',
                hostname
            )

        return ReportEmail(
                recipient=recipient,
                subject=f'Vulnerability Scan Results for {hostname}',
                plain_content=plain,
                html_content=document.to_html()
            )


VULN_SCAN_REPORT_CONFIG_OPTIONS = get_config_options(
        VulnScanReportFormat,
        VulnScanReportColumn,
        default_format='human'
    )


class VulnScanReportManager(ReportManager):

    def __init__(
                self,
                context: CliContext,
                feed_variant: VulnerabilityFeedVariant
            ):
        super().__init__(
                formats=VulnScanReportFormat,
                columns=VulnScanReportColumn,
                context=context,
                read_stdin=context.config.read_stdin,
                input_delimiter=context.config.path_separator,
                binary_input=True
            )
        self.feed_variant = feed_variant

    def _instantiate_report(
                self,
                format: ReportFormat,
                columns: List[ReportColumn],
                email_addresses: List[str],
                mailer: Optional[Mailer],
                write_headers: bool
            ) -> VulnScanReport:
        for column in columns:
            if not column.is_compatible(self.feed_variant):
                raise ValueError(
                        f'Column {column.header} is not compatible '
                        'with the current feed'
                    )
        return VulnScanReport(
                format,
                columns,
                email_addresses,
                mailer,
                write_headers
            )



import pymysql
from typing import Optional, Generator, Dict, Any

from .exceptions import WordpressDatabaseException


DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 3306
DEFAULT_USER = 'root'
DEFAULT_PREFIX = 'wp_'
DEFAULT_COLLATION = 'utf8mb4_unicode_ci'


class WordpressDatabaseConnection:

    def __init__(self, database):
        self.database = database
        try:
            self.connection = pymysql.connect(
                    host=database.server.host,
                    port=database.server.port,
                    user=database.server.user,
                    password=database.server.password,
                    database=database.name
                )
            self.set_collation(database.collation)
        except pymysql.MySQLError:
            raise WordpressDatabaseException(
                    database,
                    f'Failed to connect to database: {database.debug_string}'
                )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return

    def prefix_table(self, table: str) -> str:
        return self.database.prefix_table(table)

    def query(
                self,
                query: str,
                parameters: tuple = ()
            ) -> Generator[Dict[str, Any], None, None]:
        try:
            cursor = self.connection.cursor(pymysql.cursors.DictCursor)
            cursor.execute(query, parameters)
            for result in cursor:
                yield result
            cursor.close()
        except pymysql.MySQLError:
            raise WordpressDatabaseException(
                    self.database,
                    'Failed to execute query'
                )

    def query_literal(
                self,
                query: str
            ) -> Generator[Dict[str, Any], None, None]:
        return self.query(
                query.replace('%', '%%')
            )

    def get_column_types(
                self,
                table: str,
                prefix: bool = False
            ) -> Dict[str, str]:
        if prefix:
            table = self.prefix_table(table)
        columns = {}
        for result in self.query(f'SHOW COLUMNS FROM {table}'):
            columns[result['Field'].lower()] = result['Type']
        return columns

    def set_variable(
                self,
                variable: str,
                value: str
            ) -> None:
        self.query('SET %s = %s', (variable, value))

    def set_collation(self, collation: str) -> None:
        self.set_variable('collation_connection', collation)


class WordpressDatabaseServer:

    def __init__(
                self,
                host: str = DEFAULT_HOST,
                port: int = DEFAULT_PORT,
                user: str = DEFAULT_USER,
                password: Optional[str] = None
            ):
        self.host = host
        self.port = port
        self.user = user
        self.password = password


class WordpressDatabase:

    def __init__(
                self,
                name: str,
                server: WordpressDatabaseServer,
                prefix: str = DEFAULT_PREFIX,
                collation: str = DEFAULT_COLLATION
            ):
        self.name = name
        self.server = server
        self.prefix = prefix
        self.collation = collation
        self.debug_string = self._build_debug_string()

    def connect(self) -> WordpressDatabaseConnection:
        return WordpressDatabaseConnection(self)

    def _build_debug_string(self) -> str:
        return (
                f'{self.server.user}@{self.server.host}:'
                f'{self.server.port}/{self.name}'
            )

    def prefix_table(self, table: str) -> str:
        return self.prefix + table



import os

from enum import Enum
from typing import Optional

from ..util.io import resolve_path, get_path_components
from ..util.encoding import bytes_to_str
from .site import WordpressSite
from .exceptions import WordpressException
from .extension import Extension


class FileType(str, Enum):
    CORE = 'core'
    PLUGIN = 'plugin'
    THEME = 'theme'
    UNKNOWN = 'unknown'


class FileIdentity:

    def __init__(
                self,
                type: FileType,
                site: Optional[WordpressSite] = None,
                extension: Optional[Extension] = None,
            ):
        self.type = type
        self.site = site
        self.extension = extension

    def is_final(self) -> bool:
        return False


class GroupIdentity(FileIdentity):

    def __init__(
                self,
                type: FileType,
                path: bytes,
                site: Optional[WordpressSite] = None,
                extension: Optional[Extension] = None,
                final: bool = False
            ):
        super().__init__(type, site, extension)
        self.path = path
        self.final = final

    def is_final(self) -> bool:
        return self.final


class KnownFileIdentity(FileIdentity):

    def __init__(
                self,
                type: FileType,
                local_path: bytes,
                site: Optional[WordpressSite] = None,
                extension: Optional[Extension] = None,
            ):
        super().__init__(type, site, extension)
        self.local_path = local_path

    def is_final(self) -> bool:
        return True

    def __str__(self) -> str:
        if self.extension is None:
            software = 'WordPress'
            version = self.site.get_version()
        else:
            software = self.extension.get_name()
            version = self.extension.version
        if isinstance(version, bytes):
            version = bytes_to_str(version)
        return (
                os.fsdecode(self.local_path) +
                f' of {self.type.value} {software} ({version})'
            )


class KnownPath:

    def __init__(
                self,
                identity: Optional[FileIdentity] = None
            ):
        self.identity = identity
        self.children = {}

    def is_root(self) -> bool:
        return self.path is None

    def find_identity(self, path: bytes) -> Optional[FileIdentity]:
        node = self
        path = resolve_path(path)
        identity = None
        for component in get_path_components(path):
            if node.identity is not None:
                identity = node.identity
                if node.identity.is_final():
                    break
            try:
                node = node.children[component]
            except KeyError:
                break
        if node.identity is not None:
            return node.identity
        return identity

    def set_identity(
                self,
                path: bytes,
                identity: FileIdentity,
                resolve: bool = True
            ) -> None:
        node = self
        if resolve:
            path = resolve_path(path)
        for component in get_path_components(path):
            if component not in node.children:
                node.children[component] = KnownPath()
            node = node.children[component]
        node.identity = identity

    def __str__(self) -> str:
        if self.identity is None:
            return 'Unknown Path'
        else:
            final = self.identity.is_final()
            return f'Known Path: {self.identity.type}, {final}'

    def debug(self, indent: str = '') -> None:
        print(indent + str(self))
        for path, child in self.children.items():
            print(indent + f' {path}')
            child.debug(indent + '\t')


class FileIdentifier:

    def __init__(self):
        self.known_paths = KnownPath()

    def _identify_new_path(self, path: bytes):
        try:
            site = WordpressSite(
                        path,
                        is_child_path=True,
                        allow_io_errors=True
                    )
            core_path = site.core_path
            self.known_paths.set_identity(
                    core_path,
                    GroupIdentity(
                        type=FileType.CORE,
                        path=core_path,
                        site=site
                    )
                )
            for plugin in site.get_all_plugins():
                self.known_paths.set_identity(
                        plugin.path,
                        GroupIdentity(
                            type=FileType.PLUGIN,
                            path=plugin.path,
                            site=site,
                            extension=plugin,
                            final=True
                        ),
                        resolve=False
                    )
            for theme in site.get_themes():
                self.known_paths.set_identity(
                        theme.path,
                        GroupIdentity(
                            type=FileType.THEME,
                            path=theme.path,
                            site=site,
                            extension=theme,
                            final=True
                        ),
                        resolve=False
                    )
        except WordpressException:
            self.known_paths.set_identity(
                    path,
                    FileIdentity(
                        type=FileType.UNKNOWN
                    )
                )

    def identify(self, path: bytes, identify_new: bool = True) -> FileIdentity:
        identity = self.known_paths.find_identity(path)
        if identity is None:
            if identify_new:
                self._identify_new_path(path)
                return self.identify(path, False)
            else:
                return FileIdentity(FileType.UNKNOWN)
        elif isinstance(identity, GroupIdentity):
            local_path = os.path.relpath(path, identity.path) \
                if path != identity.path \
                else path.name
            identity = KnownFileIdentity(
                    identity.type,
                    local_path,
                    identity.site,
                    identity.extension
                )
            self.known_paths.set_identity(path, identity)
        return identity



from ...logging import LogLevel

from ..terms_management import TERMS_URL
from ..mailing_lists import EMAIL_SIGNUP_MESSAGE
from ..email import SmtpTlsMode

from .defaults import INI_DEFAULT_PATH
from .config_items import config_definitions_to_config_map

config_definitions = {
    "configuration": {
        "short_name": "c",
        "description": "Path to a configuration INI file to use.",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": INI_DEFAULT_PATH,
        "meta": {
            "accepts_file": True
        }
    },
    "banner": {
        "description": "Display the Wordfence banner in command output when "
                       "running in a TTY/terminal.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True,
        "category": "Output Control"
    },
    "license": {
        "short_name": "l",
        "description": "Specify the license to use.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": None
    },
    "email": {
        "short_name": "E",
        "description": (
                "Email address(es) to which to send reports.\n"
                "The output file will be attached to the email report "
                "when the --output-path option is used. (for supported "
                "subcommands)"
            ),
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ","
        },
        "category": "Email"
    },
    "email-from": {
        "description": "The From address to use when sending emails. If "
                       "not specified, the current username and hostname "
                       "will be used.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": None,
        "category": "Email"
    },
    "smtp-host": {
        "description": "The host name of the SMTP server to use for "
                       "sending email.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": None,
        "category": "Email"
    },
    "smtp-port": {
        "description": "The port of the SMTP server to use for sending "
                       "email.",
        "context": "ALL",
        "argument_type": "OPTION",
        "meta": {
            "value_type": int
        },
        "default": None,
        "category": "Email"
    },
    "smtp-tls-mode": {
        "description": "The SSL/TLS mode to use when communicating with "
                       f"the SMTP server. {SmtpTlsMode.NONE.value} "
                       f"disables TLS entirely. {SmtpTlsMode.SMTPS.value} "
                       "requires TLS for all communication while "
                       f"{SmtpTlsMode.STARTTLS.value} will negotiate TLS "
                       "if supported using the STARTTLS SMTP command.",
        "context": "ALL",
        "argument_type": "OPTION",
        "meta": {"valid_options": [mode.value for mode in SmtpTlsMode]},
        "default": SmtpTlsMode.STARTTLS.value,
        "category": "Email"
    },
    "smtp-user": {
        "description": "The username for authenticating with the SMTP "
                       "server.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": None,
        "category": "Email"
    },
    "smtp-password": {
        "description": "The password for authentication with the SMTP "
                       "server. This should generally be specified in an "
                       "INI file as including passwords as command line "
                       "arguments can expose them to other users on the "
                       "same system.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": None,
        "category": "Email"
    },
    "sendmail-path": {
        "description": "The path to the sendmail executable. This will be "
                       "used to send email if SMTP is not configured.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": "sendmail",
        "category": "Email"
    },
    "verbose": {
        "short_name": "v",
        "description": "Enable verbose logging. If not specified, verbose "
                       "logging will be enabled automatically if stdout is a "
                       "TTY.",
        "context": "ALL",
        "argument_type": "OPTIONAL_FLAG",
        "default": None,
        "category": "Logging"
    },
    "debug": {
        "short_name": "d",
        "description": "Enable debug logging.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False,
        "category": "Logging"
    },
    "quiet": {
        "short_name": "q",
        "description": "Suppress all output other than scan results.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False,
        "category": "Logging"
    },
    "color": {
        "description": "Enable ANSI escape sequences in output.",
        "context": "CLI",
        "argument_type": "OPTIONAL_FLAG",
        "default": None,
        "category": "Output Control"
    },
    "prefix-log-levels": {
        "description": "Prefix log messages with their respective levels. "
                       "This is enabled by default when colored output is "
                       "not enabled.",
        "context": "CLI",
        "argument_type": "OPTIONAL_FLAG",
        "default": None,
        "category": "Logging"
    },
    "log-level": {
        "short_name": "L",
        "description": "Only log messages at or above the specified level.",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": None,
        "meta": {
            "valid_options": [level.name for level in LogLevel]
        },
        "category": "Logging"
    },
    "cache-directory": {
        "description": "A path to use for cache files.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": b'~/.cache/wordfence',
        "category": "Caching",
        "meta": {
            "accepts_directory": True
        }
    },
    "cache": {
        "description": "Enable caching. Caching is enabled by default.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True,
        "category": "Caching"
    },
    "purge-cache": {
        "description": "Purge any existing values from the cache.",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False,
        "category": "Caching"
    },
    "version": {
        "description": "Display the version of Wordfence CLI.",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False
    },
    "help": {
        "short_name": "h",
        "description": "Display help information.",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False
    },
    "accept-terms": {
        "description": "Automatically accept the terms required to invoke "
                       "the specified command. The latest terms can be "
                       "viewed using the wordfence terms command  and found "
                       f"at {TERMS_URL}. {EMAIL_SIGNUP_MESSAGE}",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False
    },
    # Hidden options
    "check-for-update": {
        "description": "Whether or not to run the update check.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True,
        "hidden": True
    },
    "noc1-url": {
        "description": "URL to use for accessing the NOC1 API.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": None,
        "hidden": True
    },
    "wfi-url": {
        "description": "Base URL for accessing the Wordfence Intelligence API",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": None,
        "hidden": True
    },
}

config_map = config_definitions_to_config_map(config_definitions)



import logging
from typing import Optional
from enum import IntEnum
from dataclasses import dataclass

from .formatting import ConfigurableFormatter

DEFAULT_LOGGER_NAME = 'wordfence'
VERBOSE = 15


class LogLevel(IntEnum):
    DEBUG = logging.DEBUG
    VERBOSE = VERBOSE
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


logging.addLevelName(LogLevel.VERBOSE.value, LogLevel.VERBOSE.name)

logging.basicConfig(format='%(message)s')
log = logging.getLogger(DEFAULT_LOGGER_NAME)
root_log = logging.getLogger()

initial_handler: Optional[logging.Handler] = None


def remove_initial_handler() -> None:
    global initial_handler
    if initial_handler is not None:
        return
    initial_handler = root_log.handlers[0]
    root_log.removeHandler(initial_handler)


def restore_initial_handler(error_if_not_set: bool = False) -> None:
    global initial_handler
    if initial_handler is None:
        if error_if_not_set:
            raise ValueError("Unknown initial handler")
        return
    root_log.addHandler(initial_handler)
    initial_handler = None


def set_log_format(colored: bool = False, prefixed: bool = False) -> None:
    for handler in root_log.handlers:
        handler.setFormatter(
                ConfigurableFormatter(colored, prefixed)
            )


@dataclass
class LogSettings:
    level: LogLevel = LogLevel.WARNING
    colored: bool = False
    prefixed: bool = False

    def apply(self) -> None:
        log.setLevel(self.level)
        set_log_format(
                colored=self.colored,
                prefixed=self.prefixed
            )



from ..library import LibraryNotAvailableException


class VectorscanException(Exception):
    pass


class VectorscanLibraryNotAvailableException(
            VectorscanException,
            LibraryNotAvailableException
        ):
    pass


try:
    from .bindings import *  # noqa: F401, F403
    AVAILABLE = True
except VectorscanLibraryNotAvailableException:
    AVAILABLE = False



import sys
from typing import Optional, Any, Callable, Set, Union

from ..version import __version__, __version_name__
from ..util import pcre, vectorscan
from ..util.text import yes_no
from ..api import noc1, intelligence
from ..util.caching import Cache, CacheDirectory, RuntimeCache, \
        InvalidCachedValueException, CacheException
from ..util.input import has_terminal_input, has_terminal_output
from ..util.io import resolve_path
from ..api.licensing import License, LicenseRequiredException, \
        LicenseSpecific, to_license
from ..logging import log, LogLevel, LogSettings
from .config.config import Config
from .email import Mailer


class CliContext:

    def __init__(
                self,
                config: Config,
                cacheable_types: Set[str],
                helper,
                allows_color: bool
            ):
        self.config = config
        self.cacheable_types = cacheable_types
        self.set_up_cache(self.config.cache_directory)
        self.helper = helper
        self.allows_color = allows_color
        self._license = None
        self._noc1_client = None
        self._terms_update_hooks = []
        self._license_update_hooks = []
        self._wfi_client = None
        self._mailer = None
        self.configurer = None
        self._log_settings = None

    def get_log_level(self) -> LogLevel:
        if self.config.log_level is not None:
            return LogLevel[self.config.log_level]
        elif self.config.quiet:
            return LogLevel.CRITICAL
        elif self.config.debug:
            return LogLevel.DEBUG
        elif self.config.verbose or (
                    self.config.verbose is None
                    and sys.stdout is not None and sys.stdout.isatty()
                ):
            return LogLevel.VERBOSE
        else:
            return LogLevel.INFO

    def get_log_settings(self) -> LogSettings:
        if self._log_settings is None:
            prefixed = not self.allows_color \
                if self.config.prefix_log_levels is None \
                else self.config.prefix_log_levels
            self._log_settings = LogSettings(
                    level=self.get_log_level(),
                    colored=self.allows_color,
                    prefixed=prefixed
                )
        return self._log_settings

    def initialize_logging(self) -> None:
        settings = self.get_log_settings()
        settings.apply()

    def set_up_cache(self, directory: bytes) -> None:
        cache = self._initialize_cache(directory)
        cache.add_filter(self.filter_cache_entry)
        self.cache = cache

    def _initialize_cache(self, directory: bytes) -> Cache:
        if self.config.cache:
            try:
                return CacheDirectory(
                        resolve_path(directory),
                        self.cacheable_types
                    )
            except CacheException as exception:
                log.warning(
                        'Failed to initialize directory cache: '
                        + str(exception)
                    )
        return RuntimeCache()

    def register_terms_update_hook(
                self,
                callable: Callable[[bool, License], None]
            ) -> None:
        self._terms_update_hooks.append(callable)

    def register_license_update_hook(
                self,
                callable: Callable[[License], None]
            ) -> None:
        self._license_update_hooks.append(callable)

    def get_license(self) -> Optional[License]:
        if self._license is None and self.config.license is not None:
            self._license = License(self.config.license)
        return self._license

    def require_license(self) -> License:
        license = self.get_license()
        if license is None:
            raise LicenseRequiredException()
        return license

    def filter_cache_entry(self, value: Any) -> Any:
        if isinstance(value, LicenseSpecific):
            if not value.is_compatible_with_license(self.require_license()):
                raise InvalidCachedValueException(
                        'Incompatible license'
                    )
        return value

    def create_noc1_client(
                self,
                license: Optional[Union[License, str]] = None,
                use_hooks: bool = True
            ) -> noc1.Client:
        license = to_license(license)
        client = noc1.Client(
                license,
                self.config.noc1_url
            )
        if use_hooks:
            for hook in self._terms_update_hooks:
                client.register_terms_update_hook(hook)
            for hook in self._license_update_hooks:
                client.register_license_update_hook(hook)
        return client

    def get_noc1_client(self) -> noc1.Client:
        if self._noc1_client is None:
            self._noc1_client = self.create_noc1_client(
                    self.require_license(),
                    use_hooks=True
                )
        return self._noc1_client

    def get_wfi_client(self) -> intelligence.Client:
        if self._wfi_client is None:
            self._wfi_client = intelligence.Client(
                    self.config.wfi_url
                )
        return self._wfi_client

    def get_mailer(self) -> Mailer:
        if self._mailer is None:
            self._mailer = Mailer(self.config)
        return self._mailer

    def has_pcre(self) -> bool:
        return pcre.AVAILABLE

    def has_vectorscan(self) -> bool:
        return vectorscan.AVAILABLE

    def display_version(self) -> None:
        if __version_name__ is None:
            name_suffix = ''
        else:
            name_suffix = f' "{__version_name__}"'
        print(f"Wordfence CLI {__version__}{name_suffix}")
        has_pcre = self.has_pcre()
        pcre_support_text = yes_no(has_pcre)
        if has_pcre:
            jit_support_text = yes_no(pcre.HAS_JIT_SUPPORT)
            pcre_support_text += (
                    f" - PCRE Version: {pcre.VERSION}"
                    f" (JIT Supported: {jit_support_text})"
                )
        print(f'PCRE Supported: {pcre_support_text}')
        has_vectorscan = self.has_vectorscan()
        vectorscan_support_text = yes_no(has_vectorscan)
        if has_vectorscan:
            vectorscan_support_text += (
                    f' - Version: {vectorscan.VERSION} (API Version: '
                    f'{vectorscan.API_VERSION})'
                )
        print(f'Vectorscan Supported: {vectorscan_support_text}')

    def has_terminal_output(self) -> bool:
        return has_terminal_output()

    def has_terminal_input(self) -> bool:
        return has_terminal_input()

    def requires_input(self, option: Optional[bool]) -> bool:
        return (
                option is True or
                (option is None and self.has_terminal_input())
            )

    def clean_up(self) -> None:
        if self._mailer is not None:
            self._mailer.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        self.clean_up()



import os
from typing import Optional, Dict

from ..util.io import PathProperties
from .extension import Extension, ExtensionLoader


PLUGIN_HEADER_FIELDS = {
        'Name': 'Plugin Name',
        'PluginURI': 'Plugin URI',
        'Version': 'Version',
        'Description': 'Description',
        'Author': 'Author',
        'AuthorURI': 'Author URI',
        'TextDomain': 'Test Domain',
        'DomainPath': 'Domain Path',
        'Network': 'Network',
        'RequiresWP': 'Requires at least',
        'RequiresPHP': 'Requires PHP',
        '_sitewide': 'Site Wide Only'
    }


class Plugin(Extension):
    pass


class PluginLoader(ExtensionLoader):

    def __init__(self, directory: str, allow_io_errors: bool = False):
        super().__init__(
                'plugin',
                directory=directory,
                header_fields=PLUGIN_HEADER_FIELDS,
                allow_io_errors=allow_io_errors
            )

    def _initialize_extension(
                self,
                slug: str,
                version: Optional[str],
                header: Dict[str, str],
                path: bytes
            ):
        return Plugin(
                slug=slug,
                version=version,
                header=header,
                path=path
            )

    def _has_php_extension(self, properties: PathProperties) -> bool:
        return properties.extension == b'.php'

    def _process_entry(self, entry: os.DirEntry) -> Optional[Plugin]:
        # Ignore dot files
        if entry.name.find(b'.') == 0:
            return None
        if entry.is_file():
            path_properties = PathProperties(entry.path)
            if self._has_php_extension(path_properties):
                return self.load(
                        os.fsdecode(path_properties.stem),
                        entry.path,
                    )
        elif entry.is_dir():
            slug = os.fsdecode(entry.name)
            for child in os.scandir(entry.path):
                if child.is_file():
                    child_path = os.path.join(entry.path, child.name)
                    if self._has_php_extension(PathProperties(child_path)):
                        plugin = self.load(
                                slug,
                                child_path,
                                base_path=entry.path
                            )
                        if plugin is not None:
                            return plugin
        return None



from urllib.parse import urlparse, urlunparse, parse_qs, urlencode


class Url:

    def __init__(self, url: str):
        self._parts = urlparse(url)._asdict()

    def get_hostname(self) -> str:
        return self._parts['netloc']

    def get_query(self) -> str:
        return self._parts['query']

    def set_query(self, query: str) -> None:
        self._parts['query'] = query

    def set_query_parameter(self, key, value) -> None:
        parameters = parse_qs(self.get_query())
        parameters[key] = value
        self.set_query(urlencode(parameters, doseq=True))

    def __str__(self) -> str:
        return urlunparse(self._parts.values())



from ..subcommands import SubcommandDefinition, UsageExample
from ..config.typing import ConfigDefinitions
from .reporting import REMEDIATION_REPORT_CONFIG_OPTIONS

config_definitions: ConfigDefinitions = {
    "read-stdin": {
        "description": "Read paths from stdin. If not specified, paths will "
                       "automatically be read from stdin when input is not "
                       "from a TTY.",
        "context": "ALL",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "path-separator": {
        "short_name": "s",
        "description": "Separator used to delimit paths when reading from "
                       "stdin. Defaults to the null byte.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": "AA==",
        "default_type": "base64"
    },
    **REMEDIATION_REPORT_CONFIG_OPTIONS,
    "output-unremediated": {
        "short_name": "u",
        "description": "Only include unremediated paths in the output.",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False,
        "category": "Output Control"
    },
    "require-path": {
        "description": "When enabled, invoking the remediate command without "
                       "specifying at least one path will trigger an error. "
                       "This is the default behavior when running in a "
                       "terminal.",
        "context": "CLI",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
}

examples = [
    UsageExample(
        'Restore the original contents of a plugin file',
        'wordfence remediate /var/www/html/wp-content/plugins/hello.php'
    ),
    UsageExample(
        'Restore all files in a theme directory and output the results to a '
        'CSV file',
        'wordfence remediate --output-format csv --output-path '
        '/tmp/wfcli-remediation-results.csv --output-headers '
        '/var/www/html/wp-content/themes/twentytwentythree'
    ),
    UsageExample(
        'Automatically detect and remediate malware under /var/www/wordpress',
        'wordfence malware-scan --output-columns filename -m null-delimited '
        '/var/www/wordpress | wordfence remediate'
    )
]

definition = SubcommandDefinition(
    name='remediate',
    usage='[OPTIONS] [PATH]...',
    description='Remediate malware by restoring the content of known files',
    long_description='Known files will be overwritten with their original '
                     'content from the WordPress.org repo. Any intentional '
                     'modifications will be lost if files are remediated. '
                     'Performing a backup of existing files prior to '
                     'remediation is recommended.',
    config_definitions=config_definitions,
    config_section='REMEDIATE',
    cacheable_types=set(),
    examples=examples
)



from typing import Optional


def bytes_to_str(value: Optional[bytes]) -> Optional[str]:
    if value is None:
        return None
    return value.decode('latin1', 'replace')


def str_to_bytes(value: Optional[str]) -> Optional[bytes]:
    if value is None:
        return None
    return value.encode('latin1', 'replace')



from argparse import ArgumentParser
from types import SimpleNamespace
from typing import Any, Dict, Optional

from .config_items import ConfigItemDefinition, Context


class Config(SimpleNamespace):

    def __init__(
                self,
                definitions,
                parser: ArgumentParser,
                subcommand: Optional[str],
                ini_path: Optional[str] = None
            ):
        super().__init__()
        self._definitions = definitions
        self._parser = parser
        self.subcommand = subcommand
        self.ini_path = ini_path
        self.trailing_arguments = None
        self.defaulted_options = set()
        self.sources = {}

    def values(self) -> Dict[str, Any]:
        result: Dict[str, Any] = dict()
        for prop, value in vars(self).items():
            if (prop.startswith('_') or callable(value) or
                    isinstance(value, classmethod)):
                continue
            result[prop] = value
        return result

    def get(self, property_name, default=None) -> Any:
        return getattr(self, property_name, default)

    def define(self, property_name) -> ConfigItemDefinition:
        return self._definitions[property_name]

    def has_ini_file(self) -> bool:
        return self.ini_path is not None

    def display_help(self) -> None:
        self._parser.print_help()
        print()

    def is_specified(self, option: str) -> bool:
        return option not in self.defaulted_options

    def is_from_cli(self, option: str) -> bool:
        return option in self.sources \
            and self.sources[option] is Context.CLI



from ctypes import c_int

from ..library import LibraryNotAvailableException


class PcreException(Exception):
    pass


class PcreLibraryNotAvailableException(
            PcreException,
            LibraryNotAvailableException
        ):
    pass


PCRE_CASELESS = 0x00000001
PCRE_DEFAULT_MATCH_LIMIT = 1000000
PCRE_DEFAULT_MATCH_LIMIT_RECURSION = 100000


class PcreOptions:

    def __init__(
                self,
                caseless: bool = False,
                match_limit: int = PCRE_DEFAULT_MATCH_LIMIT,
                match_limit_recursion: int = PCRE_DEFAULT_MATCH_LIMIT_RECURSION
            ):
        self.caseless = caseless
        self.match_limit = match_limit
        self.match_limit_recursion = match_limit_recursion
        self._compilation_options = None

    def _get_compilation_options(self) -> c_int:
        if self._compilation_options is None:
            options = 0
            if self.caseless:
                options |= PCRE_CASELESS
            self._compilation_options = c_int(options)
        return self._compilation_options


PCRE_DEFAULT_OPTIONS = PcreOptions()


try:
    from .bindings import *  # noqa: F401, F403
    AVAILABLE = True
except PcreLibraryNotAvailableException:
    AVAILABLE = False



import fcntl
import os
import errno
from typing import Optional, IO, TextIO, Generator, Iterable, List, Union, \
    Callable, Set, BinaryIO
from enum import Enum, IntEnum
from collections import deque


SYMLINK_IO_ERRORS = {
        errno.ENOENT,  # File not found
        errno.ELOOP  # Too many levels of symbolic links
    }


class IoException(Exception):
    pass


class StreamReader:

    def __init__(
                self,
                stream: Union[TextIO, BinaryIO],
                delimiter: Union[str, bytes],
                chunk_size: int = 1024,
                binary: bool = False
            ):
        self.stream = stream
        self.delimiter = delimiter
        self.chunk_size = chunk_size
        self.binary = binary
        self._initialize_buffer()
        self._end_of_stream = False
        self._empty = self._initialize_empty_string()

    def _initialize_empty_string(self) -> Union[str, bytes]:
        return b'' if self.binary else ''

    def _initialize_buffer(self) -> None:
        self._buffer = self._initialize_empty_string()

    def read_entry(self) -> Optional[Union[str, bytes]]:
        while True:
            index = self._buffer.find(self.delimiter)
            if index != -1:
                entry = self._buffer[:index]
                self._buffer = self._buffer[index + 1:]
                return entry
            elif not self._end_of_stream:
                read = os.read(
                        self.stream.fileno(),
                        self.chunk_size
                    )
                if not self.binary:
                    read = read.decode(self.stream.encoding)
                if read == self._empty:
                    self._end_of_stream = True
                self._buffer += read
            else:
                break
        if len(self._buffer) > 0:
            path = self._buffer
            self._initialize_buffer()
            return path
        else:
            return None

    def read_all_entries(self) -> Generator[Union[str, bytes], None, None]:
        while (entry := self.read_entry()) is not None:
            yield entry


class LockType(IntEnum):
    EXCLUSIVE = fcntl.LOCK_EX
    SHARED = fcntl.LOCK_SH


class FileLock:

    def __init__(self, file: IO, lock_type: LockType = LockType.EXCLUSIVE):
        self.file = file
        self.lock_type = lock_type

    def _operate_lock(self, action: int):
        fcntl.flock(self.file, action)

    def __enter__(self):
        self._operate_lock(self.lock_type.value)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._operate_lock(fcntl.LOCK_UN)


def resolve_path(path: bytes) -> bytes:
    """ Resolve a path to a normalized, absolute path """
    return os.path.abspath(os.path.expanduser(path))


def resolve_parent_path(path: bytes) -> bytes:
    return resolve_path(os.path.dirname(path))


def get_path_components(path: bytes) -> bytes:
    components = []
    while len(path) > 0:
        path, current = os.path.split(path)
        if len(current) > 0:
            components.append(current)
        else:
            break
    components.reverse()
    return components


# A memory-optimized tree-set implementation for paths
class PathSet:

    def __init__(self):
        self.children = {}

    def _get_components(self, path: bytes) -> Iterable[bytes]:
        return split_path(path)

    def add(self, path: bytes) -> None:
        components = self._get_components(path)
        node = self.children
        for component in components:
            try:
                node = node[component]
            except KeyError:
                child = {}
                node[component] = child
                node = child

    def contains(self, path: bytes) -> bool:
        components = self._get_components(path)
        node = self.children
        for component in components:
            try:
                node = node[component]
            except KeyError:
                return False
        return True

    def __contains__(self, path) -> bool:
        if not isinstance(path, bytes):
            return False
        return self.contains(path)


DEFAULT_CREATE_MODE = 0o700


def ensure_directory_is_writable(
            path: bytes,
            create_mode: int = DEFAULT_CREATE_MODE
        ) -> bytes:
    """ Ensure that the specified directory is writable, """
    """ creating it and parent directories as needed. Note """
    """ that the checks here are not atomic; this assumes """
    """ that nothing else is modifying the filesystem which """
    """ is not guaranteed. """
    path = resolve_path(path)
    if os.path.exists(path):
        if os.path.isdir(path):
            if not os.access(path, os.W_OK):
                raise IoException(f'Directory at {path} is not writable')
            if not os.access(path, os.X_OK):
                raise IoException(f'Directory at {path} is not executable')
            return path
        else:
            raise IoException(f'Path {path} exists and is not a directory')
    try:
        os.makedirs(path, mode=create_mode, exist_ok=True)
    except OSError:
        raise IoException(f'Failed to create directory at {path}')
    return path


def ensure_file_is_writable(
            path: bytes,
            create_mode: int = 0o700
        ) -> bytes:
    path = resolve_path(path)
    if os.path.exists(path):
        if not os.path.isfile(path):
            raise IoException(f'Path {path} already exists, but is not a file')
        if not os.access(path, os.W_OK):
            raise IoException(f'File at {path} is not writable')
        return path
    else:
        return ensure_directory_is_writable(os.path.dirname(path))


class PathType(Enum):
    FILE = 'file',
    DIRECTORY = 'directory',
    LINK = 'link'


def get_path_type(path: bytes) -> PathType:
    if os.path.islink(path):
        return PathType.LINK
    elif os.path.isdir(path):
        return PathType.DIRECTORY
    else:
        return PathType.FILE


def is_same_file(path: bytes, other: bytes) -> bool:
    type = get_path_type(path)
    other_type = get_path_type(other)
    if type is not other_type:
        return False
    return os.path.samefile(path, other)


def is_symlink_error(error: OSError) -> bool:
    return error.errno in SYMLINK_IO_ERRORS


def is_symlink_loop(
            path: bytes,
            parents: Optional[Union[Iterable[bytes], PathSet]] = None
        ) -> bool:
    realpath = os.path.realpath(path)
    try:
        if is_same_file(path, realpath):
            return True
    except OSError as error:
        if error.errno == errno.ENOENT:
            return False
        if error.errno == errno.ELOOP:
            return True
        raise
    if parents is not None:
        if isinstance(parents, PathSet):
            if realpath in parents:
                return True
        else:
            for parent in parents:
                if realpath == parent:
                    return True
    return False


def is_symlink_and_loop(
            path: bytes,
            parents: Optional[Union[Iterable[bytes], PathSet]] = None
        ) -> bool:
    try:
        if not os.path.islink(path):
            return False
    except OSError:
        return False
    return is_symlink_loop(path, parents)


def get_all_parents(path: bytes) -> List[bytes]:
    parents = [path]
    while len(path) > 1:
        path = os.path.dirname(path)
        parents.append(path)
    return parents


def split_path(path: bytes) -> Iterable[bytes]:
    components = deque()
    while len(path) > 1:
        components.appendleft(os.path.basename(path))
        path = os.path.dirname(path)
    return components


def populate_parents(
            path: bytes,
            parents: Optional[Set[bytes]] = None
        ) -> Set[str]:
    if parents is None:
        parents = set()
    else:
        parents = parents.copy()
    parents.update(get_all_parents(path))
    return parents


def iterate_files(
            path: bytes,
            parents: Optional[Set[bytes]] = None,
            loop_callback: Optional[Callable[[bytes], None]] = None
        ) -> Generator[str, None, None]:
    parents = populate_parents(path, parents)
    if is_symlink_and_loop(path, parents):
        if loop_callback is not None:
            loop_callback(path)
        return
    contents = os.scandir(path)
    for item in contents:
        if is_symlink_and_loop(item.path, parents):
            if loop_callback is not None:
                loop_callback(item.path)
            continue
        if item.is_dir():
            yield from iterate_files(item.path, parents, loop_callback)
        else:
            yield item.path


def get_umask() -> int:
    current = os.umask(0)
    os.umask(current)
    return current


def umask_mode(mode: int) -> int:
    umask = get_umask()
    return mode & ~umask


def chmod_with_umask(path: bytes, mode: int = 0o666) -> int:
    mode = umask_mode(mode)
    os.chmod(path, mode)


class PathProperties:

    def __init__(self, path: bytes):
        self.path = path
        self.directory, self.basename = os.path.split(self.path)
        self.stem, self.extension = os.path.splitext(self.basename)

    def has_extension(self) -> bool:
        return len(self.extension) > 0
